"""
Módulo de Inicialización de Poblaciones para Algoritmos Genéticos.

Este módulo proporciona funciones para crear la población inicial de individuos
(cromosomas) para un algoritmo genético. Las funciones están diseñadas para ser
flexibles y eficientes, utilizando Numba para la optimización cuando es aplicable.

Funciones de Inicialización Disponibles:
    - init_random_uniform:
        Genera una población con genes de valor real o entero muestreados de una
        distribución uniforme, especificado mediante el parámetro `dtype`.
    - init_random_binary:
        Genera una población con genes binarios (0 o 1).
    - init_random_permutation:
        Genera una población donde cada individuo es una permutación de enteros
        (útil para problemas de ordenamiento).
"""

from typing import Any, List, Type, Union

import numpy as np
from numba import njit
from numpy.typing import NDArray

from .exceptions import (
    InitializationError,
    RangeError,
    TypeValidationError,
    validate_parameter,
)


@njit(cache=True)
def _generate_float_uniform_population_impl(
    pop_size: int, chromosome_length: int, low: float, high: float
) -> List[NDArray[np.float64]]:
    """
    Generar población uniforme de flotantes.

    Implementación Numba para generar población uniforme de flotantes.

    Args:
        pop_size: Número de individuos en la población.
        chromosome_length: Número de genes en cada cromosoma (individuo).
        low: Límite inferior del rango para los valores de los genes.
        high: Límite superior del rango para los valores de los genes.
              (Exclusivo para flotantes, inclusivo para enteros).

    Returns:
        Una lista de individuos (arrays de NumPy) del `dtype` especificado.
    """
    population = []
    for _ in range(pop_size):
        # np.random.uniform genera valores en [low, high)
        individual = np.random.uniform(low, high, size=chromosome_length)
        population.append(individual)  # Por defecto es float64
    return population


@njit(cache=True)
def _generate_integer_uniform_population_impl(
    pop_size: int, chromosome_length: int, low: int, high: int
) -> List[NDArray[np.int_]]:  # Numba puede usar np.int_ o np.int64
    """
    Generar población uniforme de enteros.

    Implementación Numba para generar población uniforme de enteros.

    Args:
        pop_size: Número de individuos en la población.
        chromosome_length: Número de genes en cada cromosoma (individuo).
        low: Límite inferior del rango para los valores de los genes.
        high: Límite superior del rango para los valores de los genes.
              (Exclusivo para flotantes, inclusivo para enteros).

    Returns:
        Una lista de individuos (arrays de NumPy) del `dtype` especificado.
    """
    population = []
    # np.random.randint(low, high_exclusive, size=...)
    # Para rango [low, high] inclusivo, el segundo parámetro debe ser high + 1
    for _ in range(pop_size):
        individual = np.random.randint(low, high + 1, size=chromosome_length)
        population.append(
            individual.astype(np.int_)
        )  # Asegurar tipo int_ (plataforma-dependiente int)
    return population


@njit(cache=True)
def _generate_binary_population_impl(
    pop_size: int, chromosome_length: int, p_one: float
) -> List[NDArray[np.int_]]:
    """
    Generar población binaria.

    Implementación Numba para generar población binaria.

    Args:
        pop_size: Número de individuos en la población.
        chromosome_length: Número de genes en cada cromosoma (individuo).
        p_one: Probabilidad de que un gen sea 1.

    Returns:
        Una lista de individuos (arrays de NumPy) del `dtype` especificado.
    """
    population = []
    for _ in range(pop_size):
        individual = np.random.random(size=chromosome_length) < p_one
        population.append(
            individual.astype(np.int_)
        )  # Convertir booleano a int (0 o 1)
    return population


@njit(cache=True)
def _generate_permutation_population_impl(
    pop_size: int, chromosome_length: int
) -> List[NDArray[np.int_]]:
    """
    Generar población de permutaciones de enteros.

    Implementación Numba para generar población de permutaciones de enteros.

    Args:
        pop_size: Número de individuos en la población.
        chromosome_length: Número de genes en cada cromosoma (individuo).

    Returns:
        Una lista de individuos (arrays de NumPy) del `dtype` especificado.
    """
    population = []
    base_array = np.arange(
        chromosome_length, dtype=np.int_
    )  # Especificar dtype para claridad
    for _ in range(pop_size):
        individual = np.random.permutation(base_array)
        population.append(individual)  # Ya es del tipo correcto (np.int_)
    return population


def init_random_uniform(
    pop_size: int,
    chromosome_length: int,
    low: Union[float, int],
    high: Union[float, int],
    dtype: Type[
        Union[
            np.float32,
            np.float64,
            np.int32,
            np.int64,
            np.int_,
        ]
    ] = np.float64,  # type: ignore
    **kwargs: Any,
) -> List[NDArray[Any]]:
    """
    Función para crear una población con genes con distribución uniforme.

    Genera una población con genes de valor real o entero muestreados de una
    distribución uniforme. El tipo de dato (flotante o entero) y su precisión
    se determinan por el parámetro `dtype`.

    - Si `dtype` es flotante (e.g., `np.float64`, `np.float32`):
        Los genes se muestrean de U(low, high), donde `high` es exclusivo.
    - Si `dtype` es entero (e.g., `np.int_`, `np.int32`, `np.int64`):
        Los genes se muestrean uniformemente del rango `[low, high]`,
        donde `high` es inclusivo.

    Args:
        pop_size: Número de individuos en la población.
        chromosome_length: Número de genes en cada cromosoma (individuo).
        low: Límite inferior del rango para los valores de los genes.
        high: Límite superior del rango para los valores de los genes.
              (Exclusivo para flotantes, inclusivo para enteros).
        dtype: Tipo de dato NumPy deseado para los genes.
               Debe ser un tipo flotante (e.g., `np.float64`, `np.float32`)
               o entero (e.g., `np.int_`, `np.int32`, `np.int64`).
               Por defecto es `np.float64`.
        **kwargs: Argumentos adicionales (no utilizados).

    Returns:
        Una lista de individuos (arrays de NumPy) del `dtype` especificado.

    Raises:
        TypeValidationError: Si los tipos de los parámetros son incorrectos,
                             o `dtype` no es un tipo NumPy numérico soportado.
        RangeError: Si `pop_size` o `chromosome_length` no son positivos,
                    o si los rangos `low`, `high` son inválidos para el `dtype`.
        InitializationError: Si ocurre un error inesperado durante la inicialización.

    Example:
        >>> # Población de flotantes (por defecto dtype=np.float64)
        >>> pop_float = init_random_uniform(5, 3, 0.0, 1.0)
        >>> len(pop_float)
        5
        >>> pop_float[0].shape
        (3,)
        >>> pop_float[0].dtype
        dtype('float64')

        >>> # Población de enteros de 32 bits
        >>> pop_int32 = init_random_uniform(5, 4, 0, 10, dtype=np.int32)
        >>> len(pop_int32)
        5
        >>> pop_int32[0].shape
        (4,)
        >>> pop_int32[0].dtype
        dtype('int32')
        >>> np.all(pop_int32[0] >= 0) and np.all(pop_int32[0] <= 10)
        True
    """
    validate_parameter(pop_size, "pop_size", int, min_val=1)
    validate_parameter(chromosome_length, "chromosome_length", int, min_val=1)
    validate_parameter(low, "low", (float, int))  # type: ignore
    validate_parameter(high, "high", (float, int))  # type: ignore

    # Validación de dtype
    valid_float_dtypes = [np.float32, np.float64]
    valid_int_dtypes = [np.int32, np.int64, np.int_]

    is_float_type = dtype in valid_float_dtypes
    is_int_type = dtype in valid_int_dtypes

    if not (is_float_type or is_int_type):
        allowed_dtypes_str = ", ".join(
            [t.__name__ for t in valid_float_dtypes + valid_int_dtypes]
        )
        raise TypeValidationError(
            "dtype",
            str(dtype),
            f"uno de los tipos NumPy soportados: {allowed_dtypes_str}",
        )

    # Validación de rango low/high según dtype
    if is_float_type and float(low) >= float(high):
        raise RangeError(
            "low",
            low,
            max_val=float(high),
            inclusive=False,  # type: ignore
            details="""Para tipos flotantes,
            'low' debe ser estrictamente menor que 'high'.""",
        )
    if is_int_type and int(low) > int(high):
        raise RangeError(
            "low",
            low,
            max_val=int(high),
            inclusive=True,  # type: ignore
            details="""Para tipos enteros,
            'low' no debe ser mayor que 'high'.""",
        )

    try:
        if is_float_type:
            # _generate_float_uniform_population_impl devuelve float64
            population_raw = _generate_float_uniform_population_impl(
                pop_size, chromosome_length, float(low), float(high)
            )
            if dtype == np.float32:  # Solo castear si se pide float32 específicamente
                population = [ind.astype(np.float32) for ind in population_raw]
            else:  # Es np.float64
                population = population_raw
        else:  # is_int_type
            # _generate_integer_uniform_population_impl devuelve np.int_
            population_raw = _generate_integer_uniform_population_impl(
                pop_size, chromosome_length, int(low), int(high)
            )
            # Castear al dtype entero específico
            # si es necesario y diferente de np.int_ default
            if population_raw[0].dtype != np.dtype(dtype):
                population = [
                    ind.astype(dtype) for ind in population_raw
                ]  # type: ignore
            else:
                population = population_raw
    except Exception as e:
        raise InitializationError(
            f"Error durante la inicialización uniforme: {str(e)}",
            details={  # type: ignore
                "pop_size": pop_size,
                "chromosome_length": chromosome_length,
                "low": low,
                "high": high,
                "dtype": str(dtype),
                "original_error": str(e),
            },
        ) from e
    return population  # type: ignore


def init_random_binary(
    pop_size: int,
    chromosome_length: int,
    p_one: float = 0.5,
    **kwargs: Any,
) -> List[NDArray[np.int_]]:
    """
    Genera una población con genes binarios (0 o 1).

    Cada gen se establece en 1 con probabilidad `p_one`
    y en 0 con probabilidad `1 - p_one`.
    Los genes son de tipo `np.int_`.

    Args:
        pop_size: Número de individuos en la población.
        chromosome_length: Número de genes en cada cromosoma.
        p_one: Probabilidad de que un gen individual sea 1. Por defecto es 0.5.
        **kwargs: Argumentos adicionales (no utilizados).

    Returns:
        Una lista de individuos (arrays de NumPy de tipo `np.int_`), donde cada
        individuo representa un cromosoma binario.

    Raises:
        TypeValidationError: Si los tipos de los parámetros son incorrectos.
        RangeError: Si `pop_size` o `chromosome_length` no son positivos,
                    o si `p_one` no está en el rango [0, 1].
        InitializationError: Si ocurre un error inesperado durante la inicialización.

    Example:
        >>> pop_bin = init_random_binary(10, 8)
        >>> len(pop_bin)
        10
        >>> pop_bin[0].shape
        (8,)
        >>> pop_bin[0].dtype # np.int_ es int32 o int64 según la plataforma
        dtype('int...')
        >>> np.all((pop_bin[0] == 0) | (pop_bin[0] == 1))
        True
    """
    validate_parameter(pop_size, "pop_size", int, min_val=1)
    validate_parameter(chromosome_length, "chromosome_length", int, min_val=1)
    validate_parameter(p_one, "p_one", float, min_val=0.0, max_val=1.0)  # type: ignore

    try:
        # _generate_binary_population_impl devuelve una lista de NDArray[np.int_]
        population = _generate_binary_population_impl(
            pop_size, chromosome_length, p_one
        )
    except Exception as e:
        raise InitializationError(
            f"Error durante la inicialización binaria: {str(e)}",
            details={  # type: ignore
                "pop_size": pop_size,
                "chromosome_length": chromosome_length,
                "p_one": p_one,
                "original_error": str(e),
            },
        ) from e
    return population


def init_random_permutation(
    pop_size: int,
    chromosome_length: int,
    **kwargs: Any,
) -> List[NDArray[np.int_]]:
    """
    Genera una población donde cada individuo es una permutación de enteros.

    Cada cromosoma será una permutación de los enteros
    `0, 1, ..., chromosome_length - 1`.
    Este método está diseñado para problemas de optimización combinatoria (e.g., TSP)
    donde los individuos representan un orden o secuencia de elementos discretos.
    Los genes son de tipo `np.int_`.

    Args:
        pop_size: Número de individuos en la población.
        chromosome_length: Longitud de la permutación (número de elementos a permutar).
        **kwargs: Argumentos adicionales (no utilizados).

    Returns:
        Una lista de individuos (arrays de NumPy de tipo `np.int_`), donde cada
        individuo es una permutación de enteros.

    Raises:
        TypeValidationError: Si los tipos de los parámetros son incorrectos.
        RangeError: Si `pop_size` o `chromosome_length` no son positivos.
        InitializationError: Si ocurre un error inesperado durante la inicialización.

    Example:
        >>> pop_perm = init_random_permutation(5, 4)
        >>> len(pop_perm)
        5
        >>> pop_perm[0].shape
        (4,)
        >>> pop_perm[0].dtype # np.int_ es int32 o int64 según la plataforma
        dtype('int...')
        >>> import numpy as np
        >>> np.array_equal(np.sort(pop_perm[0]), np.arange(4))
        True
    """
    validate_parameter(pop_size, "pop_size", int, min_val=1)
    validate_parameter(chromosome_length, "chromosome_length", int, min_val=1)

    try:
        population = _generate_permutation_population_impl(pop_size, chromosome_length)
    except Exception as e:
        raise InitializationError(
            f"Error durante la inicialización de permutaciones: {str(e)}",
            details={  # type: ignore
                "pop_size": pop_size,
                "chromosome_length": chromosome_length,
                "original_error": str(e),
            },
        ) from e
    return population



"""PySAG: Una librería simple de Algoritmos Genéticos en Python."""

from . import crossover, initialization, mutation, selection
from .exceptions import (
    GeneticAlgorithmError,
    InitializationError,
    MutationError,
    ParameterError,
    RangeError,
    SelectionError,
    TypeValidationError,
)
from .ga import GA

__version__ = "0.1.2"

__all__ = [
    "GA",
    "crossover",
    "initialization",
    "mutation",
    "selection",
    "GeneticAlgorithmError",
    "InitializationError",
    "MutationError",
    "ParameterError",
    "RangeError",
    "SelectionError",
    "TypeValidationError",
]



"""
Módulo de Selección para Algoritmos Genéticos.

Este módulo proporciona varios métodos de selección que pueden ser utilizados
en algoritmos genéticos. Todas las funciones están optimizadas con Numba
para mejorar el rendimiento y cuentan con validación de entradas y manejo
de excepciones robusto.

Funciones de Selección Disponibles:
    - selection_roulette_wheel:
        Selección proporcional al fitness (ruleta).
    - selection_tournament:
        Selección por torneo entre un subconjunto de individuos.
    - selection_rank:
        Selección basada en el ranking de fitness de los individuos.
    - selection_stochastic_universal_sampling (SUS):
        Variante de la ruleta que reduce el azar en la asignación de cupos.
    - selection_random:
        Selección aleatoria simple de individuos.
"""

from typing import Any, List

import numpy as np
from numba import njit
from numpy.typing import NDArray

from .exceptions import (
    RangeError,
    SelectionError,
    TypeValidationError,
    validate_parameter,
)


def _validate_selection_inputs(
    population: List[NDArray[np.float64]],
    fitness_values: NDArray[np.float64],
    num_parents: int,
    population_size_min: int = 1,
) -> None:
    """
    Valida los parámetros comunes de las funciones de selección.

    Args:
        population: Lista de individuos (arrays de NumPy).
        fitness_values: Array de NumPy con los valores de fitness.
        num_parents: Número de padres a seleccionar.
        population_size_min: Tamaño mínimo permitido para la población.

    Raises:
        TypeValidationError: Si los tipos de los parámetros son incorrectos.
        ValueError: Si la población está vacía, las longitudes no coinciden,
                    o num_parents es inválido.
        RangeError: Si num_parents está fuera del rango permitido.
    """
    if not isinstance(population, list):
        raise TypeValidationError("population", population, List[NDArray[np.float64]])
    if not all(isinstance(ind, np.ndarray) for ind in population):
        raise TypeValidationError(
            "population",
            "algunos elementos no son np.ndarray",
            List[NDArray[np.float64]],
        )
    validate_parameter(fitness_values, "fitness_values", NDArray[np.float64])
    validate_parameter(num_parents, "num_parents", int)

    if not population:
        raise ValueError("La población no puede estar vacía.")
    if len(population) < population_size_min:
        raise ValueError(
            f"La población debe tener al menos {population_size_min} individuos."
        )
    if len(population) != len(fitness_values):
        raise ValueError(
            "La longitud de la población y de los valores de fitness debe ser la misma."
        )
    if num_parents <= 0:
        raise RangeError("num_parents", num_parents, min_val=1)


@njit(cache=True)
def _roulette_wheel_selection_impl(
    fitness_values: NDArray[np.float64], num_parents: int
) -> NDArray[np.int_]:
    """
    Implementación Numba para la selección por ruleta.

    Selecciona 'num_parents' individuos de la población de forma aleatoria,
    con probabilidad proporcional a sus valores de fitness.
    Optimizada con decorador @njit de numba para mejorar el rendimiento.

    Args:
        fitness_values: Array de NumPy con los valores de fitness.
        num_parents: Número de padres a seleccionar.

    Returns:
        Lista de índices de los individuos seleccionados.

    Raises:
        ValueError: Si todos los valores de fitness son 0.
    """
    fitness_sum = np.sum(fitness_values)
    if fitness_sum == 0:
        # Si todos los fitness son 0, la probabilidad es uniforme
        probabilities = np.ones(len(fitness_values), dtype=np.float64) / len(
            fitness_values
        )
    else:
        probabilities = fitness_values / fitness_sum

    # np.random.choice no es soportado directamente en Numba
    # con probabilidades no uniformes
    # de la misma manera que en NumPy puro para devolver índices.
    # Implementamos una versión simple de la selección por ruleta:
    selected_indices = np.empty(num_parents, dtype=np.int_)
    cumulative_prob = np.cumsum(probabilities)
    for i in range(num_parents):
        random_val = np.random.random()
        for j in range(len(cumulative_prob)):
            if random_val <= cumulative_prob[j]:
                selected_indices[i] = j
                break
    return selected_indices


def selection_roulette_wheel(
    population: List[NDArray[np.float64]],
    fitness_values: NDArray[np.float64],
    num_parents: int,
    **kwargs: Any,
) -> List[NDArray[np.float64]]:
    """
    Selección por ruleta.

    Los individuos son seleccionados con una probabilidad proporcional a su fitness.
    Se asume que todos los valores de fitness son no negativos.

    Args:
        population: Lista de individuos (arrays de NumPy) que componen la población.
        fitness_values: Array de NumPy con los valores de fitness de cada individuo.
                        Deben ser no negativos.
        num_parents: Número de padres a seleccionar.
        **kwargs: Argumentos adicionales (no utilizados en esta función).

    Returns:
        Lista de individuos seleccionados como padres.

    Raises:
        TypeValidationError: Si los tipos de los parámetros son incorrectos.
        ValueError: Si la población está vacía, las longitudes no coinciden,
                    o si algún valor de fitness es negativo.
        RangeError: Si num_parents es inválido.
        SelectionError: Si ocurre un error inesperado durante la selección.

    Example:
        >>> import numpy as np
        >>> pop = [np.array([1,2]), np.array([3,4]), np.array([5,6]), np.array([7,8])]
        >>> fit = np.array([0.1, 0.4, 0.3, 0.2])
        >>> parents = selection_roulette_wheel(pop, fit, 2)
        >>> len(parents)
        2
        >>> isinstance(parents[0], np.ndarray)
        True
    """
    _validate_selection_inputs(population, fitness_values, num_parents)
    if np.any(fitness_values < 0):
        str_err = "Todos los valores de fitness deben ser"
        str_err += " no negativos para la selección por ruleta."
        raise ValueError(str_err)

    try:
        selected_indices = _roulette_wheel_selection_impl(fitness_values, num_parents)
        selected_parents = [population[i].copy() for i in selected_indices]
    except Exception as e:
        raise SelectionError(
            f"Error durante la selección por ruleta: {str(e)}",
            details={
                "population_size": len(population),
                "num_parents": num_parents,
                "original_error": str(e),
            },
        ) from e
    return selected_parents


@njit(cache=True)
def _tournament_selection_impl(
    fitness_values: NDArray[np.float64],
    num_parents: int,
    tournament_size: int,
    population_size: int,
) -> NDArray[np.int_]:
    """
    Implementación Numba para la selección por torneo.

    Selecciona 'num_parents' individuos de la población de forma aleatoria,
    con probabilidad proporcional a sus valores de fitness.
    optimizada con decorador @njit de numba para mejorar el rendimiento.

    Args:
        fitness_values: Array de NumPy con los valores de fitness.
        num_parents: Número de padres a seleccionar.
        tournament_size: Tamaño del torneo.
        population_size: Tamaño de la población.

    Returns:
        Lista de índices de los individuos seleccionados.

    Raises:
        ValueError: Si todos los valores de fitness son 0.
    """
    selected_indices = np.empty(num_parents, dtype=np.int_)
    population_indices = np.arange(population_size)

    for i in range(num_parents):
        if (
            tournament_size > population_size
        ):  # Numba no permite que el error se eleve desde aquí
            # Esta condición se manejará en la función principal
            pass

        contender_indices = np.random.choice(
            population_indices, size=tournament_size, replace=False
        )

        winner_in_tournament = -1
        best_fitness_in_tournament = -np.inf

        for contender_idx in contender_indices:
            if fitness_values[contender_idx] > best_fitness_in_tournament:
                best_fitness_in_tournament = fitness_values[contender_idx]
                winner_in_tournament = contender_idx
        selected_indices[i] = winner_in_tournament
    return selected_indices


def selection_tournament(
    population: List[NDArray[np.float64]],
    fitness_values: NDArray[np.float64],
    num_parents: int,
    tournament_size: int = 3,
    **kwargs: Any,
) -> List[NDArray[np.float64]]:
    """
    Selección por torneo.

    Se eligen 'tournament_size' individuos al azar, y el mejor de ellos (mayor fitness)
    se convierte en padre. Este proceso se repite 'num_parents' veces.

    Args:
        population: Lista de individuos (arrays de NumPy).
        fitness_values: Array de NumPy con los valores de fitness.
        num_parents: Número de padres a seleccionar.
        tournament_size: Número de individuos que participan en cada torneo.
                         Por defecto es 3.
        **kwargs: Argumentos adicionales (no utilizados en esta función).

    Returns:
        Lista de individuos seleccionados como padres.

    Raises:
        TypeValidationError: Si los tipos de los parámetros son incorrectos.
        ValueError: Si la población tiene menos individuos que `tournament_size`.
        RangeError: Si num_parents o tournament_size son inválidos.
        SelectionError: Si ocurre un error inesperado durante la selección.

    Example:
        >>> import numpy as np
        >>> pop = [np.array([1,2]), np.array([3,4]), np.array([5,6]), np.array([7,8])]
        >>> fit = np.array([10, 40, 30, 20])
        >>> parents = selection_tournament(pop, fit, 2, tournament_size=2)
        >>> len(parents)
        2
    """
    _validate_selection_inputs(
        population, fitness_values, num_parents, population_size_min=1
    )
    validate_parameter(tournament_size, "tournament_size", int, min_val=1)

    if tournament_size > len(population):
        raise ValueError(
            f"El tamaño del torneo ({tournament_size}) no puede ser mayor "
            f"que el tamaño de la población ({len(population)})."
        )
    if tournament_size == 0:
        raise RangeError(
            "tournament_size",
            tournament_size,
            min_val=1,
            details="Tournament size no puede ser cero.",
        )

    try:
        selected_indices = _tournament_selection_impl(
            fitness_values, num_parents, tournament_size, len(population)
        )
        selected_parents = [population[i].copy() for i in selected_indices]
    except Exception as e:
        raise SelectionError(
            f"Error durante la selección por torneo: {str(e)}",
            details={
                "population_size": len(population),
                "num_parents": num_parents,
                "tournament_size": tournament_size,
                "original_error": str(e),
            },
        ) from e
    return selected_parents


@njit(cache=True)
def _rank_selection_impl(
    fitness_values: NDArray[np.float64], num_parents: int
) -> NDArray[np.int_]:
    """
    Implementación Numba para la selección por ranking.

    Optimizada con decorador @njit de numba para mejorar el rendimiento.

    Args:
        fitness_values: Array de NumPy con los valores de fitness.
        num_parents: Número de padres a seleccionar.

    Returns:
        Lista de índices de los individuos seleccionados.
    """
    # argsort devuelve los índices que ordenarían el array
    sorted_indices_fitness = np.argsort(fitness_values)

    # Crear ranks (el mejor fitness obtiene el mayor rango)
    ranks = np.empty_like(sorted_indices_fitness, dtype=np.float64)
    ranks[sorted_indices_fitness] = np.arange(1, len(fitness_values) + 1)

    rank_sum = np.sum(ranks)
    if rank_sum == 0:  # Todos los rangos son 0 (imposible con arange(1, N+1))
        # o la población es 0 (manejado antes).
        # Esto es más una guarda por si los rangos fueran negativos o cero.
        probabilities = np.ones(len(fitness_values), dtype=np.float64) / len(
            fitness_values
        )
    else:
        probabilities = ranks / rank_sum

    # Implementación de np.random.choice con probabilidades para Numba
    selected_indices = np.empty(num_parents, dtype=np.int_)
    cumulative_prob = np.cumsum(probabilities)
    for i in range(num_parents):
        random_val = np.random.random()
        # Encuentra el primer índice j tal que random_val <= cumulative_prob[j]
        # Esto asegura que los individuos con mayor probabilidad (mayor rango)
        # tengan más chance de ser seleccionados.
        # Dado que las probabilidades se basan en 'ranks', que a su vez se derivan
        # de 'sorted_indices_fitness', la selección final de 'j' aquí
        # es un índice directo en la población original.
        selected_idx = 0
        for j in range(len(cumulative_prob)):
            if random_val <= cumulative_prob[j]:
                selected_idx = j
                break
        selected_indices[i] = selected_idx

    return selected_indices


def selection_rank(
    population: List[NDArray[np.float64]],
    fitness_values: NDArray[np.float64],
    num_parents: int,
    **kwargs: Any,
) -> List[NDArray[np.float64]]:
    """
    Selección por Rango.

    Los individuos son ordenados según su fitness y se les asigna un rango.
    La probabilidad de selección es proporcional a este rango
    (mayor rango, mayor probabilidad).
    Este método puede evitar la dominancia prematura de individuos con fitness muy alto.

    Args:
        population: Lista de individuos (arrays de NumPy).
        fitness_values: Array de NumPy con los valores de fitness.
        num_parents: Número de padres a seleccionar.
        **kwargs: Argumentos adicionales (no utilizados en esta función).

    Returns:
        Lista de individuos seleccionados como padres.

    Raises:
        TypeValidationError: Si los tipos de los parámetros son incorrectos.
        ValueError: Si la población está vacía o las longitudes no coinciden.
        RangeError: Si num_parents es inválido.
        SelectionError: Si ocurre un error inesperado durante la selección.

    Example:
        >>> import numpy as np
        >>> pop = [np.array([1,2]), np.array([3,4]), np.array([5,6]), np.array([7,8])]
        >>> fit = np.array([10, 40, 30, 20]) # Ranks: 10 (1), 20 (2), 30 (3), 40 (4)
        >>> parents = selection_rank(pop, fit, 2)
        >>> len(parents)
        2
    """
    _validate_selection_inputs(population, fitness_values, num_parents)

    try:
        selected_indices = _rank_selection_impl(fitness_values, num_parents)
        selected_parents = [population[i].copy() for i in selected_indices]
    except Exception as e:
        raise SelectionError(
            f"Error durante la selección por rango: {str(e)}",
            details={
                "population_size": len(population),
                "num_parents": num_parents,
                "original_error": str(e),
            },
        ) from e
    return selected_parents


@njit(cache=True)
def _stochastic_universal_sampling_impl(
    fitness_values: NDArray[np.float64], num_parents: int
) -> NDArray[np.int_]:
    """
    Implementación Numba para Stochastic Universal Sampling (SUS).

    Selecciona 'num_parents' individuos de la población de forma aleatoria,
    con probabilidad proporcional a sus valores de fitness.
    Optimizada con decorador @njit de numba para mejorar el rendimiento.

    Args:
        fitness_values: Array de NumPy con los valores de fitness.
        num_parents: Número de padres a seleccionar.

    Returns:
        Lista de índices de los individuos seleccionados.
    """
    fitness_sum = np.sum(fitness_values)
    if fitness_sum == 0:
        # Aquí, si todos son cero, la probabilidad es uniforme.
        return np.random.randint(0, len(fitness_values), size=num_parents)

    selected_indices = np.empty(num_parents, dtype=np.int_)

    # Distancia entre punteros
    pointer_distance = fitness_sum / num_parents

    # Posición inicial del primer puntero (aleatoria entre 0 y pointer_distance)
    start_pointer = np.random.uniform(0, pointer_distance)

    pointers = (
        np.arange(num_parents, dtype=np.float64) * pointer_distance + start_pointer
    )

    cumulative_fitness = 0.0
    current_member = 0
    for i_pointer in range(num_parents):
        # Moverse a través de los miembros hasta que el puntero sea superado
        while cumulative_fitness < pointers[i_pointer]:
            if current_member >= len(fitness_values):  # salvaguarda
                current_member = len(fitness_values) - 1  # seleccionar el ultimo
                break
            cumulative_fitness += fitness_values[current_member]
            current_member += 1  # Preparar para el siguiente miembro
        selected_indices[i_pointer] = (
            current_member - 1
        )  # El miembro que superó el puntero
        if selected_indices[i_pointer] < 0:  # Asegurarse de que no sea negativo
            selected_indices[i_pointer] = 0

    return selected_indices


def selection_stochastic_universal_sampling(
    population: List[NDArray[np.float64]],
    fitness_values: NDArray[np.float64],
    num_parents: int,
    **kwargs: Any,
) -> List[NDArray[np.float64]]:
    """
    Selección por Muestreo Universal Estocástico (SUS).

    Es una técnica de selección que minimiza el factor de suerte inherente en la
    selección por ruleta. Utiliza un único giro de una "ruleta" con N punteros
    equidistantes para seleccionar N individuos. Esto asegura que el número de
    veces que un individuo es seleccionado es más cercano a su proporción esperada.
    Se asume que todos los valores de fitness son no negativos.

    Args:
        population: Lista de individuos (arrays de NumPy).
        fitness_values: Array de NumPy con los valores de fitness.
        Deben ser no negativos.
        num_parents: Número de padres a seleccionar.
        **kwargs: Argumentos adicionales (no utilizados en esta función).

    Returns:
        Lista de individuos seleccionados como padres.

    Raises:
        TypeValidationError: Si los tipos de los parámetros son incorrectos.
        ValueError: Si la población está vacía, las longitudes no coinciden,
                    o si algún valor de fitness es negativo.
        RangeError: Si num_parents es inválido.
        SelectionError: Si ocurre un error inesperado durante la selección.

    Example:
        >>> import numpy as np
        >>> pop = [np.array([1,2]), np.array([3,4]), np.array([5,6]), np.array([7,8])]
        >>> fit = np.array([0.1, 0.4, 0.3, 0.2])
        >>> parents = selection_stochastic_universal_sampling(pop, fit, 2)
        >>> len(parents)
        2
    """
    _validate_selection_inputs(population, fitness_values, num_parents)
    if np.any(fitness_values < 0):
        raise ValueError(
            "Todos los valores de fitness deben ser no negativos para SUS."
        )
    if num_parents == 0:
        pass

    try:
        selected_indices = _stochastic_universal_sampling_impl(
            fitness_values, num_parents
        )
        selected_parents = [population[i].copy() for i in selected_indices]
    except Exception as e:
        raise SelectionError(
            f"Error durante la selección SUS: {str(e)}",
            details={
                "population_size": len(population),
                "num_parents": num_parents,
                "original_error": str(e),
            },
        ) from e
    return selected_parents


@njit(cache=True)
def _random_selection_impl(population_size: int, num_parents: int) -> NDArray[np.int_]:
    """
    Implementación Numba para la selección aleatoria simple.

    Optimizada con decorador @njit de numba para mejorar el rendimiento.

    Args:
        population_size: Tamaño de la población.
        num_parents: Número de padres a seleccionar.

    Returns:
        Lista de índices de los individuos seleccionados.
    """
    return np.random.randint(0, population_size, size=num_parents)


def selection_random(
    population: List[NDArray[np.float64]],
    fitness_values: NDArray[
        np.float64
    ],  # No se usa, pero se mantiene por consistencia de firma
    num_parents: int,
    **kwargs: Any,
) -> List[NDArray[np.float64]]:
    """
    Selección Aleatoria Simple.

    Selecciona 'num_parents' individuos de la población de forma completamente
    aleatoria, con reemplazo. No considera los valores de fitness.

    Args:
        population: Lista de individuos (arrays de NumPy).
        fitness_values: Array de NumPy con los valores de fitness (no utilizado).
        num_parents: Número de padres a seleccionar.
        **kwargs: Argumentos adicionales (no utilizados en esta función).

    Returns:
        Lista de individuos seleccionados como padres.

    Raises:
        TypeValidationError: Si los tipos de los parámetros son incorrectos.
        ValueError: Si la población está vacía o las longitudes no coinciden.
        RangeError: Si num_parents es inválido.
        SelectionError: Si ocurre un error inesperado durante la selección.

    Example:
        >>> import numpy as np
        >>> pop = [np.array([1,2]), np.array([3,4]), np.array([5,6]), np.array([7,8])]
        >>> fit = np.array([0.1, 0.4, 0.3, 0.2]) # Fitness no se usa
        >>> parents = selection_random(pop, fit, 3)
        >>> len(parents)
        3
    """
    _validate_selection_inputs(population, fitness_values, num_parents)

    try:
        selected_indices = _random_selection_impl(len(population), num_parents)
        selected_parents = [population[i].copy() for i in selected_indices]
    except Exception as e:
        raise SelectionError(
            f"Error durante la selección aleatoria: {str(e)}",
            details={
                "population_size": len(population),
                "num_parents": num_parents,
                "original_error": str(e),
            },
        ) from e
    return selected_parents



"""
Módulo que implementa operadores de mutación para algoritmos genéticos.

Este módulo proporciona varias estrategias de mutación que pueden ser utilizadas
en algoritmos genéticos. Las funciones están diseñadas para ser flexibles y
eficientes, utilizando Numba para la optimización del núcleo computacional.

Operadores de Mutación Disponibles:
    - mutation_bit_flip: Invierte bits aleatorios en un individuo binario.
    - mutation_random_gene_uniform: Reemplaza genes con valores de una dist. uniforme.
    - mutation_gaussian: Añade ruido gaussiano a genes numéricos.
    - mutation_swap: Intercambia dos genes aleatorios en el individuo.
    - mutation_inversion: Invierte un segmento del cromosoma (para permutaciones).
"""

from typing import Any, Dict, Optional, TypeVar, Union

import numpy as np
from numba import njit
from numpy.typing import NDArray

from .exceptions import (
    MutationError,
    ParameterError,
    RangeError,
    TypeValidationError,
    validate_parameter,
)

# Tipo genérico para los cromosomas
ChromosomeType = TypeVar("ChromosomeType", bound=np.generic)
IndividualNDArray = NDArray[ChromosomeType]


def _validate_mutation_inputs(
    individual: IndividualNDArray,
    mutation_rate: float,
    param_name: str = "individual",
    check_empty: bool = True,
    expected_ndim: int = 1,
) -> None:
    """
    Valida los parámetros comunes de las funciones de mutación.

    Args:
        individual: Individuo a mutar.
        mutation_rate: Tasa de mutación.
        param_name: Nombre del parámetro 'individual' para mensajes de error.
        check_empty: Si es True, valida que el individuo no esté vacío.
        expected_ndim: Dimensión esperada para el array del individuo.

    Raises:
        TypeValidationError: Si el tipo del individuo es incorrecto.
        ParameterError: Si el individuo está vacío o no tiene la dimensión esperada.
        RangeError: Si `mutation_rate` está fuera del rango [0, 1].
    """
    validate_parameter(individual, param_name, np.ndarray)  # type: ignore
    validate_parameter(
        mutation_rate,
        "mutation_rate",
        (float, int),
        min_val=0.0,
        max_val=1.0,
    )  # type: ignore

    if individual.ndim != expected_ndim:
        raise ParameterError(
            f"El parámetro '{param_name}' debe ser un array {expected_ndim}D. "
            f"Dimensión encontrada: {individual.ndim}."
        )
    if check_empty and individual.size == 0:
        raise ParameterError(
            f"El parámetro '{param_name}' no puede ser un array vacío."
        )


@njit(cache=True)
def _mutation_bit_flip_impl(
    individual_copy: NDArray[np.int_], mutation_rate: float
) -> NDArray[np.int_]:
    """
    Implementación Numba: Invierte bits según la tasa de mutación por gen.

    Args:
        individual_copy: Copia del individuo a mutar.
        mutation_rate: Tasa de mutación.

    Returns:
        El individuo mutado.
    """
    for i in range(len(individual_copy)):
        if np.random.random() < mutation_rate:
            individual_copy[i] = 1 - individual_copy[i]
    return individual_copy


@njit(cache=True)
def _mutation_random_gene_uniform_impl(
    individual_copy: IndividualNDArray,
    gene_low: float,
    gene_high: float,
    mutation_rate: float,
    is_float_type: bool,
) -> IndividualNDArray:
    """
    Implementación Numba: Reemplaza genes con valor uniforme según tasa por gen.

    Args:
        individual_copy: Copia del individuo a mutar.
        gene_low: Valor mínimo para el reemplazo.
        gene_high: Valor máximo para el reemplazo.
        mutation_rate: Tasa de mutación.
        is_float_type: Si es True, el individuo es de tipo float.

    Returns:
        El individuo mutado.
    """
    for i in range(len(individual_copy)):
        if np.random.random() < mutation_rate:
            if is_float_type:
                individual_copy[i] = np.random.uniform(gene_low, gene_high)
            else:  # Es entero
                # np.random.randint es [low, high)
                # para [low, high] inclusivo, usar high + 1
                individual_copy[i] = np.random.randint(
                    int(gene_low), int(gene_high) + 1
                )
    return individual_copy


@njit(cache=True)
def _mutation_gaussian_impl(
    individual_copy: NDArray[np.float64],
    mu: float,
    sigma: float,
    mutation_rate: float,
) -> NDArray[np.float64]:
    """
    Implementación Numba: Añade ruido gaussiano según tasa por gen.

    Args:
        individual_copy: Copia del individuo a mutar.
        mu: Media de la distribución normal.
        sigma: Desviación estándar de la distribución normal.
        mutation_rate: Tasa de mutación.

    Returns:
        El individuo mutado.
    """
    for i in range(len(individual_copy)):
        if np.random.random() < mutation_rate:
            noise = np.random.normal(mu, sigma)
            individual_copy[i] += noise
    return individual_copy


@njit(cache=True)
def _mutation_swap_impl(individual_copy: IndividualNDArray) -> IndividualNDArray:
    """
    Implementación Numba: Intercambia dos genes aleatorios.

    Args:
        individual_copy: Copia del individuo a mutar.

    Returns:
        El individuo mutado.
    """
    idx1, idx2 = np.random.choice(len(individual_copy), size=2, replace=False)

    gene1 = individual_copy[idx1]
    individual_copy[idx1] = individual_copy[idx2]
    individual_copy[idx2] = gene1
    return individual_copy


@njit(cache=True)
def _mutation_inversion_impl(individual_copy: IndividualNDArray) -> IndividualNDArray:
    """
    Implementación Numba: Invierte un segmento aleatorio del cromosoma.

    Args:
        individual_copy: Copia del individuo a mutar.

    Returns:
        El individuo mutado.
    """
    idx1, idx2 = np.sort(np.random.choice(len(individual_copy), size=2, replace=False))

    segment = individual_copy[idx1 : idx2 + 1]
    individual_copy[idx1 : idx2 + 1] = segment[::-1]
    return individual_copy


def mutation_bit_flip(
    individual: NDArray[np.int_],
    mutation_rate: float = 0.01,
    **kwargs: Dict[str, Any],
) -> NDArray[np.int_]:
    """
    Invierte bits aleatorios en un individuo de representación binaria.

    Cada bit tiene una probabilidad `mutation_rate` de ser invertido.

    Args:
        individual: Individuo a mutar (array NumPy 1D de enteros, 0s y 1s).
        mutation_rate: Probabilidad de que cada bit mute. Por defecto es 0.01.
        **kwargs: Argumentos adicionales (no utilizados).

    Returns:
        Una copia del individuo con los bits posiblemente invertidos.

    Raises:
        TypeValidationError:
        Si el individuo no es un array de NumPy o no es de tipo entero.

        ParameterError:
        Si el individuo no es 1D.

        RangeError:
        Si `mutation_rate` está fuera del rango [0, 1].

        ValueError:
        Si el individuo no es binario (no contiene solo 0s y 1s).

        MutationError:
        Si ocurre un error inesperado durante la mutación.

    Example:
        >>> import numpy as np
        >>> ind = np.array([0, 1, 0, 1, 0, 0, 1, 1])
        >>> mutated = mutation_bit_flip(ind, mutation_rate=0.5)
        >>> mutated.shape == ind.shape
        True
        >>> # El resultado es probabilístico, pero algunos bits deberían cambiar.
    """
    _validate_mutation_inputs(individual, mutation_rate, check_empty=False)

    if not np.issubdtype(individual.dtype, np.integer):
        raise TypeValidationError(
            "individual.dtype", str(individual.dtype), "np.integer"
        )

    if individual.size > 0 and not np.all(np.isin(individual, [0, 1])):
        raise ValueError(
            "El individuo para 'mutation_bit_flip'"
            + " debe ser binario (contener solo 0s y 1s)."
        )

    if mutation_rate == 0.0 or individual.size == 0:
        return individual.copy()

    try:
        # La función Numba maneja la iteración y la tasa de mutación por gen.
        mutated_individual = _mutation_bit_flip_impl(individual.copy(), mutation_rate)
    except Exception as e:
        raise MutationError(
            f"Error durante la mutación de inversión de bits: {str(e)}",
            details={  # type: ignore
                "mutation_rate": mutation_rate,
                "individual_shape": individual.shape,
                "individual_dtype": str(individual.dtype),
                "original_error": str(e),
            },
        ) from e
    return mutated_individual


def mutation_random_gene_uniform(
    individual: IndividualNDArray,
    gene_low: Union[float, int],
    gene_high: Union[float, int],
    mutation_rate: float = 0.01,
    **kwargs: Dict[str, Any],
) -> IndividualNDArray:
    """
    Muta genes aleatorios reemplazándolos con un valor de una distribución uniforme.

    Cada gen tiene una probabilidad `mutation_rate` de ser mutado.
    Si el `dtype` del individuo es flotante, se usa `np.random.uniform(low, high)`.
    Si es entero, se usa `np.random.randint(low, high + 1)`.

    Args:
        individual: Individuo a mutar (array NumPy 1D).
        gene_low: Límite inferior para los nuevos valores de los genes.
        gene_high: Límite superior para los nuevos valores de los genes.
                   (Exclusivo para flotantes, inclusivo para enteros).
        mutation_rate: Probabilidad de que cada gen mute. Por defecto es 0.01.
        **kwargs: Argumentos adicionales (no utilizados).

    Returns:
        Una copia del individuo con genes posiblemente mutados.

    Raises:
        TypeValidationError: Si `individual` no es array, o si `gene_low`/`gene_high`
                             no son numéricos.
        ParameterError: Si el individuo no es 1D.
        RangeError: Si `mutation_rate` está fuera de rango, o si `gene_low > gene_high`.
        MutationError: Si ocurre un error inesperado.

    Example:
        >>> import numpy as np
        >>> ind_float = np.array([1.0, 2.0, 3.0, 4.0])
        >>> mutated_float = mutation_random_gene_uniform(ind_float, 0.0, 10.0, 0.5)
        >>> ind_int = np.array([1, 2, 3, 4])
        >>> mutated_int = mutation_random_gene_uniform(ind_int, 0, 10, 0.5)
    """
    _validate_mutation_inputs(individual, mutation_rate, check_empty=False)
    validate_parameter(gene_low, "gene_low", (float, int))  # type: ignore
    validate_parameter(gene_high, "gene_high", (float, int))  # type: ignore

    is_float_type = np.issubdtype(individual.dtype, np.floating)
    is_int_type = np.issubdtype(individual.dtype, np.integer)

    if not (is_float_type or is_int_type):
        raise TypeValidationError(
            "individual.dtype",
            str(individual.dtype),
            "un tipo numérico (flotante o entero)",
        )

    if float(gene_low) > float(
        gene_high
    ):  # Comparar como flotantes para evitar problemas de tipo
        raise RangeError(
            "gene_low",
            gene_low,
            max_val=gene_high,  # type: ignore
            details="'gene_low' no puede ser mayor que 'gene_high'.",
        )

    if mutation_rate == 0.0 or individual.size == 0:
        return individual.copy()

    try:
        mutated_individual = _mutation_random_gene_uniform_impl(
            individual.copy(),
            float(gene_low),
            float(gene_high),
            mutation_rate,
            is_float_type,
        )
    except Exception as e:
        raise MutationError(
            f"Error durante la mutación uniforme de gen aleatorio: {str(e)}",
            details={  # type: ignore
                "gene_low": gene_low,
                "gene_high": gene_high,
                "mutation_rate": mutation_rate,
                "individual_shape": individual.shape,
                "individual_dtype": str(individual.dtype),
                "original_error": str(e),
            },
        ) from e
    return mutated_individual


def mutation_gaussian(
    individual: NDArray[np.float64],
    mu: float = 0.0,
    sigma: float = 1.0,
    mutation_rate: float = 0.01,
    clip_low: Optional[float] = None,
    clip_high: Optional[float] = None,
    **kwargs: Dict[str, Any],
) -> NDArray[np.float64]:
    """
    Añade ruido gaussiano a los genes del individuo.

    Cada gen tiene una probabilidad `mutation_rate` de ser mutado.
    Si muta, se le suma un valor muestreado de N(mu, sigma).
    Opcionalmente, los valores pueden ser recortados a un rango [clip_low, clip_high].

    Args:
        individual: Individuo a mutar (array NumPy 1D de flotantes).
        mu: Media de la distribución normal para el ruido. Por defecto es 0.0.
        sigma: Desviación estándar de la distribución normal. Debe ser no negativo.
               Por defecto es 1.0.
        mutation_rate: Probabilidad de que cada gen mute. Por defecto es 0.01.
        clip_low: Límite inferior opcional para recortar los valores mutados.
        clip_high: Límite superior opcional para recortar los valores mutados.
        **kwargs: Argumentos adicionales (no utilizados).

    Returns:
        Una copia del individuo con ruido gaussiano posiblemente añadido y recortado.

    Raises:
        TypeValidationError: Si `individual` no es array de flotantes, o `mu`/`sigma`
                             no son numéricos.
        ParameterError: Si el individuo no es 1D.
        RangeError: Si `mutation_rate` está fuera de rango, `sigma` es negativo,
                    o `clip_low > clip_high`.
        MutationError: Si ocurre un error inesperado.

    Example:
        >>> import numpy as np
        >>> ind = np.array([1.0, 2.0, 3.0, 4.0])
        >>> mutated = mutation_gaussian(
            ind,
            sigma=0.1,
            mutation_rate=0.5,
            clip_low=0.0,
            clip_high=5.0,
        )
    """
    _validate_mutation_inputs(individual, mutation_rate, check_empty=False)
    if not np.issubdtype(individual.dtype, np.floating):
        raise TypeValidationError(
            "individual.dtype", str(individual.dtype), "un tipo flotante de NumPy"
        )

    validate_parameter(mu, "mu", (float, int))  # type: ignore
    validate_parameter(sigma, "sigma", (float, int), min_val=0.0)  # type: ignore
    if clip_low is not None:
        validate_parameter(clip_low, "clip_low", (float, int))  # type: ignore
    if clip_high is not None:
        validate_parameter(clip_high, "clip_high", (float, int))  # type: ignore

    if clip_low is not None and clip_high is not None and clip_low > clip_high:
        raise RangeError(
            "clip_low",
            clip_low,
            max_val=clip_high,  # type: ignore
            details="'clip_low' no puede ser mayor que 'clip_high'.",
        )

    if mutation_rate == 0.0 or individual.size == 0:
        return individual.copy()

    try:
        mutated_individual = _mutation_gaussian_impl(
            individual.copy().astype(np.float64), float(mu), float(sigma), mutation_rate
        )
        if clip_low is not None or clip_high is not None:
            np.clip(mutated_individual, clip_low, clip_high, out=mutated_individual)

    except Exception as e:
        raise MutationError(
            f"Error durante la mutación gaussiana: {str(e)}",
            details={  # type: ignore
                "mu": mu,
                "sigma": sigma,
                "mutation_rate": mutation_rate,
                "clip_low": clip_low,
                "clip_high": clip_high,
                "individual_shape": individual.shape,
                "individual_dtype": str(individual.dtype),
                "original_error": str(e),
            },
        ) from e
    return mutated_individual


def mutation_swap(
    individual: IndividualNDArray,
    mutation_rate: float = 0.01,
    **kwargs: Dict[str, Any],
) -> IndividualNDArray:
    """
    Intercambia dos genes aleatorios en el individuo.

    La operación de intercambio (swap) ocurre en el individuo completo con
    una probabilidad `mutation_rate`. Si ocurre, se eligen dos genes
    distintos al azar y se intercambian sus posiciones.

    Args:
        individual: Individuo a mutar (array NumPy 1D).
        mutation_rate: Probabilidad de que la operación de intercambio ocurra.
                       Por defecto es 0.01.
        **kwargs: Argumentos adicionales (no utilizados).

    Returns:
        Una copia del individuo, con dos genes posiblemente intercambiados.

    Raises:
        TypeValidationError: Si `individual` no es un array de NumPy.
        ParameterError: Si el individuo no es 1D o tiene menos de 2 elementos
                        (necesario para un intercambio).
        RangeError: Si `mutation_rate` está fuera del rango [0, 1].
        MutationError: Si ocurre un error inesperado.

    Example:
        >>> import numpy as np
        >>> ind = np.array([1, 2, 3, 4, 5])
        >>> # Para asegurar que el swap ocurra para el ejemplo:
        >>> mutated = mutation_swap(ind, mutation_rate=1.0)
        >>> len(mutated) == len(ind) and set(mutated) == (
            set(ind) and not np.array_equal(mutated, ind)
        )
        True
    """
    _validate_mutation_inputs(individual, mutation_rate, check_empty=True)
    if individual.size < 2:
        return individual.copy()

    mutated_individual = individual.copy()
    if np.random.random() < mutation_rate:
        try:
            mutated_individual = _mutation_swap_impl(mutated_individual)
        except Exception as e:  # Captura errores de Numba o NumPy dentro de _impl
            raise MutationError(
                f"Error durante la mutación de intercambio: {str(e)}",
                details={  # type: ignore
                    "mutation_rate": mutation_rate,
                    "individual_shape": individual.shape,
                    "individual_dtype": str(individual.dtype),
                    "original_error": str(e),
                },
            ) from e
    return mutated_individual


def mutation_inversion(
    individual: IndividualNDArray,
    mutation_rate: float = 0.01,
    **kwargs: Dict[str, Any],
) -> IndividualNDArray:
    """
    Invierte un segmento aleatorio del cromosoma.

    Esta mutación es comúnmente usada para representaciones de permutación.
    La operación de inversión ocurre en el individuo completo con una
    probabilidad `mutation_rate`. Si ocurre, se seleccionan dos puntos
    aleatorios y el segmento entre ellos (inclusive) se invierte.

    Args:
        individual: Individuo a mutar (array NumPy 1D).
        mutation_rate: Probabilidad de que la operación de inversión ocurra.
                       Por defecto es 0.01.
        **kwargs: Argumentos adicionales (no utilizados).

    Returns:
        Una copia del individuo, con un segmento posiblemente invertido.

    Raises:
        TypeValidationError: Si `individual` no es un array de NumPy.
        ParameterError: Si el individuo no es 1D o tiene menos de 2 elementos
                        (necesario para una inversión significativa).
        RangeError: Si `mutation_rate` está fuera del rango [0, 1].
        MutationError: Si ocurre un error inesperado.

    Example:
        >>> import numpy as np
        >>> ind = np.array([1, 2, 3, 4, 5, 6])
        >>> # Para asegurar que la inversión ocurra para el ejemplo:
        >>> mutated = mutation_inversion(ind, mutation_rate=1.0)
        >>> len(mutated) == len(ind) and set(mutated) == set(ind)
        True
        >>> # Es probable que el orden cambie si la longitud es >= 2
    """
    _validate_mutation_inputs(individual, mutation_rate, check_empty=True)
    if individual.size < 2:
        return individual.copy()

    mutated_individual = individual.copy()
    if np.random.random() < mutation_rate:
        try:
            mutated_individual = _mutation_inversion_impl(mutated_individual)
        except Exception as e:
            raise MutationError(
                f"Error durante la mutación de inversión: {str(e)}",
                details={  # type: ignore
                    "mutation_rate": mutation_rate,
                    "individual_shape": individual.shape,
                    "individual_dtype": str(individual.dtype),
                    "original_error": str(e),
                },
            ) from e
    return mutated_individual



"""
Módulo que implementa operadores de cruce para algoritmos genéticos.

Este módulo proporciona varias estrategias de cruce que pueden ser utilizadas
en algoritmos genéticos. Las funciones están diseñadas para ser flexibles y
eficientes, utilizando Numba para la optimización del núcleo computacional.

Operadores de Cruce Disponibles:
    - crossover_single_point: Cruce de un solo punto.
    - crossover_two_points: Cruce de dos puntos.
    - crossover_uniform: Cruce uniforme, gen a gen.
    - crossover_arithmetic: Cruce aritmético para valores numéricos.
    - crossover_order_ox1: Cruce de orden (OX1) para permutaciones.
"""

from typing import Any, Tuple, TypeVar

import numpy as np
from numba import njit
from numpy.typing import NDArray

from .exceptions import (
    CrossoverError,
    ParameterError,
    TypeValidationError,
    validate_parameter,
)

# Tipo genérico para los cromosomas, permitiendo flexibilidad (float, int, bool)
ChromosomeType = TypeVar("ChromosomeType", bound=np.generic)
ParentNDArray = NDArray[ChromosomeType]


def _validate_parents(
    parent1: ParentNDArray, parent2: ParentNDArray, min_length: int = 1
) -> None:
    """
    Valida los dos padres para operaciones de cruce.

    Args:
        parent1: Primer padre.
        parent2: Segundo padre.
        min_length: Longitud mínima requerida para los cromosomas.

    Raises:
        TypeValidationError: Si los padres no son arrays de NumPy.
        ParameterError: Si los padres no tienen la misma forma, no son 1D,
                        o su longitud es menor que `min_length`.

    """
    validate_parameter(parent1, "parent1", np.ndarray)  # type: ignore
    validate_parameter(parent2, "parent2", np.ndarray)  # type: ignore

    if parent1.shape != parent2.shape:
        raise ParameterError(
            f"Los padres deben tener la misma forma. "
            f"P1 tiene forma {parent1.shape}, P2 tiene forma {parent2.shape}."
        )
    if parent1.ndim != 1:
        raise ParameterError(
            f"Los padres deben ser arrays 1D. Dimensión encontrada: {parent1.ndim}."
        )
    if len(parent1) < min_length:
        raise ParameterError(
            f"La longitud de los padres debe ser al menos {min_length}. "
            f"Longitud encontrada: {len(parent1)}."
        )
    if parent1.dtype != parent2.dtype:
        # Advertencia o error opcional. NumPy manejará la promoción de tipos,
        # pero puede ser inesperado. Por ahora, permitimos diferentes dtypes numéricos.
        # Para permutaciones, se esperarán enteros.
        pass


@njit(cache=True)
def _crossover_single_point_impl(
    p1: ParentNDArray, p2: ParentNDArray
) -> Tuple[ParentNDArray, ParentNDArray]:
    """
    Implementación Numba para cruce de un punto.

    Args:
        p1: Primer padre.
        p2: Segundo padre.

    Returns:
        Una tupla (child1, child2) con los dos descendientes.

    """
    point = np.random.randint(1, len(p1))  # Punto de cruce entre 1 y len-1

    child1_parts = (p1[:point], p2[point:])
    child2_parts = (p2[:point], p1[point:])

    # np.concatenate es soportado por Numba y maneja dtypes.
    return np.concatenate(child1_parts), np.concatenate(child2_parts)


@njit(cache=True)
def _crossover_two_points_impl(
    p1: ParentNDArray, p2: ParentNDArray
) -> Tuple[ParentNDArray, ParentNDArray]:
    """
    Implementación Numba para cruce de dos puntos.

    Args:
        p1: Primer padre.
        p2: Segundo padre.

    Returns:
        Una tupla (child1, child2) con los dos descendientes.
    """
    # Generar dos puntos distintos, point1 < point2
    point1 = np.random.randint(1, len(p1) - 1)  # de 1 a len-2
    point2 = np.random.randint(point1 + 1, len(p1))  # de point1+1 a len-1

    child1_parts = (p1[:point1], p2[point1:point2], p1[point2:])
    child2_parts = (p2[:point1], p1[point1:point2], p2[point2:])

    return np.concatenate(child1_parts), np.concatenate(child2_parts)


@njit(cache=True)
def _crossover_uniform_impl(
    p1: ParentNDArray, p2: ParentNDArray, mix_prob: float
) -> Tuple[ParentNDArray, ParentNDArray]:
    """
    Implementación Numba para cruce uniforme.

    Args:
        p1: Primer padre.
        p2: Segundo padre.
        mix_prob: Probabilidad de intercambiar genes entre padres.

    Returns:
        Una tupla (child1, child2) con los dos descendientes.
    """
    child1 = p1.copy()
    child2 = p2.copy()

    for i in range(len(p1)):
        if np.random.random() < mix_prob:
            # Intercambiar genes en esta posición
            temp = child1[i]
            child1[i] = child2[i]
            child2[i] = temp

    c1 = p1.copy()
    c2 = p2.copy()
    for i in range(len(p1)):
        if np.random.random() < mix_prob:
            gene1 = c1[i]
            c1[i] = c2[i]
            c2[i] = gene1
    return c1, c2


@njit(cache=True)
def _crossover_arithmetic_impl(
    p1: NDArray[np.float64], p2: NDArray[np.float64], alpha: float
) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    """
    Implementación Numba para cruce aritmético.

    Args:
        p1: Primer padre.
        p2: Segundo padre.
        alpha: Factor de cruce.

    Returns:
        Una tupla (child1, child2) con los dos descendientes.
    """
    child1 = alpha * p1 + (1.0 - alpha) * p2
    child2 = (1.0 - alpha) * p1 + alpha * p2
    return child1, child2


@njit(cache=True)
def _crossover_order_ox1_impl(
    p1: NDArray[np.int_], p2: NDArray[np.int_]
) -> Tuple[NDArray[np.int_], NDArray[np.int_]]:
    """
    Implementación Numba para cruce de orden OX1.

    Args:
        p1: Primer padre.
        p2: Segundo padre.

    Returns:
        Una tupla (child1, child2) con los dos descendientes.

    """
    size = len(p1)
    child1 = np.empty_like(p1)
    child2 = np.empty_like(p2)

    # Paso 1: Seleccionar dos puntos de cruce
    start, end = np.sort(np.random.choice(np.arange(size + 1), size=2, replace=False))

    # Paso 2: Copiar el segmento del primer padre
    # al primer hijo, y del segundo al segundo
    child1[start:end] = p1[start:end]
    child2[start:end] = p2[start:end]

    # Paso 3: Llenar los genes restantes para child1
    # Elementos en p1[start:end]
    p1_segment_set = set(p1[start:end])
    current_p2_idx = end
    current_child1_idx = end
    filled_count = 0
    while filled_count < size - (end - start):
        if p2[current_p2_idx % size] not in p1_segment_set:
            child1[current_child1_idx % size] = p2[current_p2_idx % size]
            current_child1_idx += 1
            filled_count += 1
        current_p2_idx += 1
        if current_p2_idx > end + size:
            break

    # Paso 4: Llenar los genes restantes para child2
    p2_segment_set = set(p2[start:end])
    current_p1_idx = end
    current_child2_idx = end
    filled_count = 0
    while filled_count < size - (end - start):
        if p1[current_p1_idx % size] not in p2_segment_set:
            child2[current_child2_idx % size] = p1[current_p1_idx % size]
            current_child2_idx += 1
            filled_count += 1
        current_p1_idx += 1
        if current_p1_idx > end + size:
            break

    return child1, child2


def crossover_single_point(
    parent1: ParentNDArray, parent2: ParentNDArray, **kwargs: Any
) -> Tuple[ParentNDArray, ParentNDArray]:
    """
    Realiza un cruce de un solo punto entre dos padres.

    Selecciona un punto de cruce aleatorio y crea dos descendientes
    intercambiando las secciones de los padres después de este punto.
    Funciona con cualquier tipo de dato en los arrays NumPy.

    Args:
        parent1: Primer padre (array NumPy 1D).
        parent2: Segundo padre (array NumPy 1D, misma forma y tipo que parent1).
        **kwargs: Argumentos adicionales (no utilizados).

    Returns:
        Una tupla (child1, child2) con los dos descendientes.

    Raises:
        TypeValidationError: Si los padres no son arrays de NumPy.
        ParameterError: Si los padres no tienen la misma forma, no son 1D,
                        o su longitud es menor que 2.
        CrossoverError: Si ocurre un error inesperado durante el cruce.

    Example:
        >>> import numpy as np
        >>> p1 = np.array([1, 2, 3, 4, 5])
        >>> p2 = np.array([6, 7, 8, 9, 10])
        >>> c1, c2 = crossover_single_point(p1, p2)
        >>> c1.shape == p1.shape and c2.shape == p2.shape
        True
        >>> (c1 == p1).all() # Puede ser False si ocurre cruce
        False

    """
    _validate_parents(parent1, parent2, min_length=1)
    if len(parent1) < 2:
        return parent1.copy(), parent2.copy()

    try:
        # Numba maneja bien dtypes numéricos y booleanos para slicing/concatenación.
        child1, child2 = _crossover_single_point_impl(parent1, parent2)
    except Exception as e:
        raise CrossoverError(
            f"Error durante el cruce de un solo punto: {str(e)}",
            details={
                "p1_dtype": str(parent1.dtype),
                "p2_dtype": str(parent2.dtype),
            },
        ) from e
    return child1, child2


def crossover_two_points(
    parent1: ParentNDArray, parent2: ParentNDArray, **kwargs: Any
) -> Tuple[ParentNDArray, ParentNDArray]:
    """
    Realiza un cruce de dos puntos entre dos padres.

    Selecciona dos puntos de cruce aleatorios y crea dos descendientes
    intercambiando la sección entre estos dos puntos.
    Funciona con cualquier tipo de dato en los arrays NumPy.

    Args:
        parent1: Primer padre (array NumPy 1D).
        parent2: Segundo padre (array NumPy 1D, misma forma y tipo que parent1).
        **kwargs: Argumentos adicionales (no utilizados).

    Returns:
        Una tupla (child1, child2) con los dos descendientes.

    Raises:
        TypeValidationError: Si los padres no son arrays de NumPy.
        ParameterError: Si los padres no tienen la misma forma, no son 1D,
                        o su longitud es menor que 3.
        CrossoverError: Si ocurre un error inesperado durante el cruce.

    Example:
        >>> import numpy as np
        >>> p1 = np.array([1, 2, 3, 4, 5, 6])
        >>> p2 = np.array([10, 20, 30, 40, 50, 60])
        >>> c1, c2 = crossover_two_points(p1, p2)
        >>> c1.shape == p1.shape and c2.shape == p2.shape
        True
    """
    _validate_parents(parent1, parent2, min_length=2)
    if len(parent1) < 3:
        return parent1.copy(), parent2.copy()

    try:
        child1, child2 = _crossover_two_points_impl(parent1, parent2)
    except Exception as e:
        raise CrossoverError(
            f"Error durante el cruce de dos puntos: {str(e)}",
            details={
                "p1_dtype": str(parent1.dtype),
                "p2_dtype": str(parent2.dtype),
            },
        ) from e
    return child1, child2


def crossover_uniform(
    parent1: ParentNDArray,
    parent2: ParentNDArray,
    mix_probability: float = 0.5,
    **kwargs: Any,
) -> Tuple[ParentNDArray, ParentNDArray]:
    """
    Realiza un cruce uniforme entre dos padres.

    Para cada gen, se decide con `mix_probability` si los genes de los padres
    se intercambian para esa posición en los hijos. Si no se intercambian,
    el hijo1 toma el gen del padre1 y el hijo2 del padre2. Si se intercambian,
    el hijo1 toma del padre2 y el hijo2 del padre1.
    Funciona con cualquier tipo de dato en los arrays NumPy.

    Args:
        parent1: Primer padre (array NumPy 1D).
        parent2: Segundo padre (array NumPy 1D, misma forma y tipo que parent1).
        mix_probability: Probabilidad de intercambiar los genes de los padres
                         para una posición dada. Debe estar entre 0 y 1.
                         Por defecto es 0.5.
        **kwargs: Argumentos adicionales (no utilizados).

    Returns:
        Una tupla (child1, child2) con los dos descendientes.

    Raises:
        TypeValidationError: Si los padres no son arrays de NumPy.
        ParameterError: Si los padres no tienen la misma forma o no son 1D.
        RangeError: Si `mix_probability` está fuera del rango [0, 1].
        CrossoverError: Si ocurre un error inesperado durante el cruce.

    Example:
        >>> import numpy as np
        >>> p1 = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
        >>> p2 = np.array([10.0, 20.0, 30.0, 40.0, 50.0])
        >>> c1, c2 = crossover_uniform(p1, p2, mix_probability=0.5)
        >>> c1.shape == p1.shape and c2.shape == p2.shape
        True

    """
    _validate_parents(parent1, parent2, min_length=1)
    validate_parameter(
        mix_probability,
        "mix_probability",
        float,
        min_val=0.0,
        max_val=1.0,
    )  # type: ignore

    if len(parent1) == 0:
        return parent1.copy(), parent2.copy()

    try:
        child1, child2 = _crossover_uniform_impl(parent1, parent2, mix_probability)
    except Exception as e:
        raise CrossoverError(
            f"Error durante el cruce uniforme: {str(e)}",
            details={  # type: ignore
                "p1_dtype": str(parent1.dtype),
                "p2_dtype": str(parent2.dtype),
                "mix_probability": mix_probability,
            },
        ) from e
    return child1, child2


def crossover_arithmetic(
    parent1: NDArray[np.float64],
    parent2: NDArray[np.float64],
    alpha: float = 0.5,
    **kwargs: Any,
) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
    """
    Realiza un cruce aritmético entre dos padres.

    Produce dos descendientes mediante una combinación lineal de los padres:
    `child1 = alpha * parent1 + (1 - alpha) * parent2`
    `child2 = (1 - alpha) * parent1 + alpha * parent2`
    Este operador es adecuado para cromosomas con genes de valor real o entero.
    Si se usan enteros, el resultado será flotante y puede requerir conversión.

    Args:
        parent1: Primer padre (array NumPy 1D de tipo numérico).
        parent2: Segundo padre (array NumPy 1D, misma forma y tipo que parent1).
        alpha: Factor de ponderación para la combinación lineal.
               Un valor común es 0.5 (promedio simple).
               Debe estar entre 0 y 1 para interpolación.
        **kwargs: Argumentos adicionales (no utilizados).

    Returns:
        Una tupla (child1, child2) con los dos descendientes (arrays de flotantes).

    Raises:
        TypeValidationError: Si los padres no son arrays de NumPy, o si sus `dtype`
                             no son numéricos (flotante o entero).
        ParameterError: Si los padres no tienen la misma forma o no son 1D.
        RangeError: Si `alpha` está fuera del rango [0, 1] (opcional, pero común).
        CrossoverError: Si ocurre un error inesperado durante el cruce.

    Example:
        >>> import numpy as np
        >>> p1 = np.array([1.0, 2.0, 3.0])
        >>> p2 = np.array([4.0, 5.0, 6.0])
        >>> c1, c2 = crossover_arithmetic(p1, p2, alpha=0.5)
        >>> c1
        array([2.5, 3.5, 4.5])
        >>> c2
        array([2.5, 3.5, 4.5])
        >>> p_int1 = np.array([1, 2, 3])
        >>> p_int2 = np.array([4, 6, 8])
        >>> c_int1, c_int2 = crossover_arithmetic(
            p_int1.astype(float), p_int2.astype(float), alpha=0.2
        )
        >>> c_int1 # Resultado es flotante
        array([3.4, 5.2, 7. ])

    """
    _validate_parents(parent1, parent2, min_length=0)

    # Validar que los dtypes son numéricos
    if not (
        np.issubdtype(parent1.dtype, np.number)
        and np.issubdtype(parent2.dtype, np.number)
    ):
        raise TypeValidationError(
            "parent1 y parent2 dtypes",
            f"P1: {parent1.dtype}, P2: {parent2.dtype}",
            "tipos numéricos de NumPy (integer o floating)",
        )
    p1_float = parent1.astype(np.float64, copy=False)
    p2_float = parent2.astype(np.float64, copy=False)

    validate_parameter(alpha, "alpha", float)

    if len(parent1) == 0:
        return p1_float.copy(), p2_float.copy()

    try:
        child1, child2 = _crossover_arithmetic_impl(p1_float, p2_float, alpha)
    except Exception as e:
        raise CrossoverError(
            f"Error durante el cruce aritmético: {str(e)}",
            details={  # type: ignore
                "p1_dtype": str(parent1.dtype),
                "p2_dtype": str(parent2.dtype),
                "alpha": alpha,
            },
        ) from e
    return child1, child2


def crossover_order_ox1(
    parent1: NDArray[np.int_], parent2: NDArray[np.int_], **kwargs: Any
) -> Tuple[NDArray[np.int_], NDArray[np.int_]]:
    """
    Realiza un cruce de orden (OX1) entre dos padres de permutación.

    Este operador es adecuado para cromosomas que representan permutaciones de enteros.
    1. Se seleccionan dos puntos de cruce aleatorios.
    2. El segmento del primer padre entre estos puntos se copia al primer hijo.
    3. Los genes restantes se toman del segundo padre en el orden en que aparecen,
       omitiendo los genes ya presentes del primer padre.
    4. El proceso se invierte para crear el segundo hijo.

    Args:
        parent1: Primer padre (array NumPy 1D de enteros, debe ser una permutación).
        parent2: Segundo padre (array NumPy 1D de enteros, misma forma y tipo,
                 debe ser una permutación de los mismos elementos).
        **kwargs: Argumentos adicionales (no utilizados).

    Returns:
        Una tupla (child1, child2) con los dos descendientes (permutaciones).

    Raises:
        TypeValidationError:
            Si los padres no son arrays de NumPy o no son de tipo entero.
        ParameterError:
            Si los padres no tienen la misma forma, no son 1D,
            o su longitud es menor que 1 (o 3 para un cruce significativo).
            También si no parecen ser permutaciones válidas (opcional).
        CrossoverError:
            Si ocurre un error inesperado durante el cruce.

    Example:
        >>> import numpy as np
        >>> p1 = np.array([1, 2, 3, 4, 5, 6, 7, 8])
        >>> p2 = np.array([8, 6, 4, 2, 7, 5, 3, 1])
        >>> # Los puntos de cruce son aleatorios, el resultado varía.
        >>> # Para un ejemplo determinista, necesitaríamos fijar los puntos de cruce.
        >>> c1, c2 = crossover_order_ox1(p1, p2)
        >>> c1.shape == p1.shape and c2.shape == p2.shape
        True
        >>> np.array_equal(np.sort(c1), np.sort(p1))
        True
        >>> np.array_equal(np.sort(c2), np.sort(p2))
        True

    """
    _validate_parents(parent1, parent2, min_length=1)

    if not (
        np.issubdtype(parent1.dtype, np.integer)
        and np.issubdtype(parent2.dtype, np.integer)
    ):
        raise TypeValidationError(
            "parent1 y parent2 dtypes para OX1",
            f"P1: {parent1.dtype}, P2: {parent2.dtype}",
            "tipos enteros de NumPy",
        )

    if len(parent1) < 2:
        return parent1.copy(), parent2.copy()

    try:
        p1_int = parent1.astype(np.int_, copy=False)
        p2_int = parent2.astype(np.int_, copy=False)
        child1, child2 = _crossover_order_ox1_impl(p1_int, p2_int)
    except Exception as e:
        raise CrossoverError(
            f"Error durante el cruce de orden OX1: {str(e)}",
            details={
                "p1_dtype": str(parent1.dtype),
                "p2_dtype": str(parent2.dtype),
            },
        ) from e
    return child1, child2



"""
Clase principal que implementa el Algoritmo Genético (AG).

Esta clase encapsula la lógica de un algoritmo genético estándar, permitiendo
la configuración de sus componentes clave como la inicialización, selección,
cruce y mutación a través de funciones personalizadas.
"""

import random
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
from numpy.typing import NDArray

from . import crossover as default_crossover
from . import initialization as default_init
from . import mutation as default_mutation
from . import selection as default_selection
from .exceptions import (
    GeneticAlgorithmError,
    InitializationError,
    validate_parameter,
)

# Tipos para mayor claridad
IndividualType = NDArray[Any]  # Un individuo es un array NumPy
PopulationType = List[IndividualType]  # La población es una lista de individuos
FitnessFunctionType = Callable[[IndividualType], float]
OperatorFunctionType = Callable[..., Any]


class GA:
    """
    Clase principal que implementa el Algoritmo Genético (AG).

    Permite configurar cada etapa del AG (inicialización, evaluación, selección,
    cruce y mutación) y ejecutar el proceso evolutivo.
    """

    def __init__(
        self,
        fitness_func: FitnessFunctionType,
        num_genes: int,
        population_size: int = 50,
        num_generations: int = 100,
        num_parents_mating: int = 10,
        initial_population_func: OperatorFunctionType = (
            default_init.init_random_uniform
        ),
        initial_pop_args: Optional[Dict[str, Any]] = None,
        selection_func: OperatorFunctionType = (
            default_selection.selection_roulette_wheel
        ),
        selection_args: Optional[Dict[str, Any]] = None,
        crossover_func: Optional[
            OperatorFunctionType
        ] = default_crossover.crossover_single_point,
        crossover_args: Optional[Dict[str, Any]] = None,
        crossover_probability: float = 0.9,
        mutation_func: OperatorFunctionType = (
            default_mutation.mutation_random_gene_uniform
        ),
        mutation_args: Optional[Dict[str, Any]] = None,
        keep_elitism_percentage: float = 0.1,
        random_seed: Optional[int] = None,
    ) -> None:
        """
        Inicializa el algoritmo genético con los parámetros dados.

        Args:
            fitness_func:
            Función que toma un individuo y devuelve su valor de fitness (float).

            num_genes:
            Número de genes en cada individuo.

            population_size:
            Tamaño de la población. Por defecto es 50.

            num_generations:
            Número de generaciones a ejecutar. Por defecto es 100.

            num_parents_mating:
            Número de individuos a seleccionar como padres
            para el cruce. Por defecto es 10.

            initial_population_func:
            Función para crear la población inicial.
            Por defecto es `init_random_uniform`.

            initial_pop_args:
            Argumentos adicionales para `initial_population_func`.

            selection_func:
            Función para seleccionar padres.
            Por defecto es `selection_roulette_wheel`.

            selection_args:
            Argumentos adicionales para `selection_func`.

            crossover_func:
            Función para realizar el cruce. Puede ser None para no usar cruce.
            Por defecto es `crossover_single_point`.

            crossover_args:
            Argumentos adicionales para `crossover_func`.

            crossover_probability:
            Probabilidad de que ocurra el cruce.
            Debe estar entre 0 y 1. Por defecto es 0.9.

            mutation_func:
            Función para realizar la mutación.
            Por defecto es `mutation_random_gene_uniform`.

            mutation_args:
            Argumentos adicionales para `mutation_func`.

            keep_elitism_percentage:
            Porcentaje de los mejores individuos de la
            generación actual que se pasan directamente
            a la siguiente. Debe estar entre 0 y 1.
            Por defecto es 0.1.

            random_seed:
            Semilla opcional para el generador de números aleatorios
            de NumPy y Python, para reproducibilidad.

        Raises:
            TypeValidationError: Si alguna de las funciones de operador no es callable.
            ParameterError: Si alguno de los parámetros numéricos está fuera de rango.
        """
        if random_seed is not None:
            random.seed(random_seed)
            np.random.seed(random_seed)

        # Validación de parámetros
        validate_parameter(
            fitness_func, "fitness_func", expected_type=Callable
        )  # type: ignore
        validate_parameter(num_genes, "num_genes", expected_type=int, min_val=1)
        validate_parameter(
            population_size, "population_size", expected_type=int, min_val=1
        )
        validate_parameter(
            num_generations, "num_generations", expected_type=int, min_val=0
        )
        validate_parameter(
            num_parents_mating, "num_parents_mating", expected_type=int, min_val=1
        )

        validate_parameter(
            initial_population_func,
            "initial_population_func",
            expected_type=Callable,
        )  # type: ignore
        validate_parameter(
            selection_func, "selection_func", expected_type=Callable
        )  # type: ignore
        if crossover_func is not None:
            validate_parameter(
                crossover_func, "crossover_func", expected_type=Callable
            )  # type: ignore
        validate_parameter(
            mutation_func, "mutation_func", expected_type=Callable
        )  # type: ignore

        validate_parameter(
            crossover_probability,
            "crossover_probability",
            expected_type=(float, int),
            min_val=0.0,
            max_val=1.0,
        )
        validate_parameter(
            keep_elitism_percentage,
            "keep_elitism_percentage",
            expected_type=(float, int),
            min_val=0.0,
            max_val=1.0,
        )

        self.fitness_func: FitnessFunctionType = fitness_func
        self.num_genes: int = num_genes
        self.population_size: int = population_size
        self.num_generations: int = num_generations
        self.num_parents_mating: int = num_parents_mating

        self.initial_population_func: OperatorFunctionType = initial_population_func
        self.initial_pop_args: Dict[str, Any] = (
            initial_pop_args if initial_pop_args is not None else {}
        )

        self.selection_func: OperatorFunctionType = selection_func
        self.selection_args: Dict[str, Any] = (
            selection_args if selection_args is not None else {}
        )

        self.crossover_func: Optional[OperatorFunctionType] = crossover_func
        self.crossover_args: Dict[str, Any] = (
            crossover_args if crossover_args is not None else {}
        )
        self.crossover_probability: float = crossover_probability

        self.mutation_func: OperatorFunctionType = mutation_func
        self.mutation_args: Dict[str, Any] = (
            mutation_args if mutation_args is not None else {}
        )

        elitism_count_float = self.population_size * keep_elitism_percentage
        self.keep_elitism_count: int = int(elitism_count_float)
        if (
            self.keep_elitism_count == 0
            and elitism_count_float > 0
            and population_size > 0
        ):
            pass  # Por ahora, se confía en la conversión a int directa.

        self.population: Optional[PopulationType] = None
        self.best_solutions_fitness: List[float] = []
        self.best_solution_overall: Optional[IndividualType] = None
        self.best_fitness_overall: float = -np.inf  # Asume maximización

    def _initialize_population(self) -> None:
        """
        Inicializa la población utilizando la función y argumentos especificados.

        Este método utiliza los parámetros `initial_population_func` y
        `initial_pop_args` definidos durante la instanciación de la clase GA.
        Modifica `self.population`.
        """
        self.population = self.initial_population_func(
            pop_size=self.population_size,
            chromosome_length=self.num_genes,
            **self.initial_pop_args,  # Aquí deben estar 'low', 'high', 'dtype'
        )
        if not isinstance(self.population, list) or not all(
            isinstance(ind, np.ndarray) for ind in self.population
        ):
            str_error = """La función de inicialización debe devolver
            una lista de arrays NumPy (individuos)."""
            raise InitializationError(str_error)  # type: ignore
        if len(self.population) != self.population_size:
            str_error = """La población inicializada
            tiene tamaño {len(self.population)}, """
            str_error += f"pero se esperaba {self.population_size}."
            raise InitializationError(str_error)  # type: ignore

    def _calculate_population_fitness(self) -> NDArray[np.float64]:
        """
        Calcula la aptitud de todos los individuos en la población actual.

        Returns:
            Un array NumPy con los valores de fitness de la población.

        Raises:
            GeneticAlgorithmError: Si la población no ha sido inicializada.
        """
        if self.population is None:
            str_error = """La población no ha sido
            inicializada antes de calcular el fitness."""
            raise GeneticAlgorithmError(str_error)  # type: ignore

        fitness_values: List[float] = []
        for individual in self.population:
            try:
                fitness_values.append(self.fitness_func(individual))
            except Exception as e:
                raise GeneticAlgorithmError(  # type: ignore
                    f"Error al calcular el fitness para el individuo {individual}: {e}",
                    details={"original_error": str(e)},  # type: ignore
                ) from e
        return np.array(fitness_values, dtype=np.float64)

    def run(self) -> Tuple[Optional[IndividualType], float]:
        """
        Ejecuta el algoritmo genético a través de las generaciones.

        Returns:
            Una tupla con la mejor solución global encontrada (Individuo) y
            su valor de fitness (float).

        Raises:
            GeneticAlgorithmError: Si ocurren errores irrecuperables durante
                                   la ejecución del AG.
        """
        if self.population is None:
            try:
                self._initialize_population()
            except InitializationError as e:
                str_error = "Error de inicialización:"
                str_error += f" {e}"
                raise GeneticAlgorithmError(str_error) from e  # type: ignore

        if self.population is None:
            str_error = "Error crítico: La población sigue siendo None "
            str_error += "después de la inicialización."
            print(str_error)
            return None, -np.inf

        for generation in range(self.num_generations):
            try:
                fitness_values = self._calculate_population_fitness()
            except GeneticAlgorithmError as e:
                print(
                    f"Error en la generación {generation + 1} calculando fitness: {e}"
                )
                # Podría decidir terminar o intentar continuar si es recuperable
                raise  # Re-elevar por ahora

            current_best_fitness_idx = np.argmax(fitness_values)
            current_best_fitness = fitness_values[current_best_fitness_idx]
            self.best_solutions_fitness.append(current_best_fitness)

            if current_best_fitness > self.best_fitness_overall:
                self.best_fitness_overall = current_best_fitness
                self.best_solution_overall = self.population[
                    current_best_fitness_idx
                ].copy()

            print(
                f"Generación {generation + 1}/{self.num_generations}: "
                f"Mejor Fitness = {self.best_fitness_overall:.4f} "
                f"(Actual: {current_best_fitness:.4f})"
            )

            # Elitismo
            elite_individuals: PopulationType = []
            if self.keep_elitism_count > 0 and self.population:
                # Ordenar por fitness descendente y tomar los 'k' mejores
                elite_indices = np.argsort(fitness_values)[-self.keep_elitism_count :]
                elite_individuals = [self.population[i].copy() for i in elite_indices]

            # Selección de Padres
            try:
                # Asegurar que num_parents_mating no exceda population_size
                actual_num_parents_mating = min(
                    self.num_parents_mating, len(self.population)
                )
                if actual_num_parents_mating < 2 and self.crossover_func is not None:
                    str_error = "Advertencia: No hay suficientes padres"
                    str_error += f" ({actual_num_parents_mating}) para el cruce. "
                    str_error += "Se omitirá el cruce."
                    print(str_error)

                parents = self.selection_func(
                    self.population,
                    fitness_values,
                    actual_num_parents_mating,  # Usar el valor ajustado
                    **self.selection_args,
                )
            except Exception as e:  # Captura errores de selección
                str_error = "Error durante la selección en la generación"
                str_error += f" {generation + 1}: {e}"
                raise GeneticAlgorithmError(str_error) from e  # type: ignore

            if not parents:  # Si la selección no devuelve padres
                print(
                    "Advertencia: La selección no devolvió padres en la generación"
                    f" {generation + 1}. "
                    "Rellenando descendencia con clones de la población actual."
                )
                # Como fallback
                # clonar de la población existente para mantener el tamaño
                parents = [
                    random.choice(self.population).copy()
                    for _ in range(actual_num_parents_mating)
                ]
                if (
                    not parents and self.population
                ):  # Si la población también está vacía, es un problema mayor
                    str_error = (
                        "La población está vacía y no se pueden seleccionar padres."
                    )
                    raise GeneticAlgorithmError(str_error)  # type: ignore

            # Generación de Descendencia (Cruce y Mutación)
            num_offspring_to_generate = self.population_size - len(elite_individuals)
            offspring_population: PopulationType = []

            # Condición para cruce: función de cruce definida y suficientes padres
            can_crossover = self.crossover_func is not None and len(parents) >= 2

            current_parent_idx = 0
            while len(offspring_population) < num_offspring_to_generate:
                if not parents:  # No hay padres de donde generar descendencia
                    if (
                        self.population
                    ):  # Si la población original aún existe, clonar de ahí
                        offspring_population.append(
                            random.choice(self.population).copy()
                        )
                        if len(offspring_population) >= num_offspring_to_generate:
                            break
                        continue
                    else:  # No hay forma de generar más individuos
                        str_error = "Error: No hay padres ni población "
                        str_error += "base para generar descendencia."
                        break

                p1 = parents[current_parent_idx % len(parents)]

                if can_crossover and random.random() < self.crossover_probability:
                    p2 = parents[
                        (current_parent_idx + 1) % len(parents)
                    ]  # Siguiente padre para cruce
                    try:
                        offspring1, offspring2 = self.crossover_func(
                            p1.copy(), p2.copy(), **self.crossover_args
                        )
                    except Exception as e:  # Captura errores de cruce
                        str_error = "Error durante el cruce en la generación"
                        str_error += f" {generation + 1}: {e}"
                        raise GeneticAlgorithmError(str_error) from e  # type: ignore

                    offspring_population.append(offspring1)
                    if len(offspring_population) < num_offspring_to_generate:
                        offspring_population.append(offspring2)
                    current_parent_idx += 2  # Avanzar dos padres
                else:  # Clonación (sin cruce o probabilidad no cumplida)
                    offspring_population.append(p1.copy())
                    current_parent_idx += 1  # Avanzar un padre

            # Mutación de la descendencia generada
            mutated_offspring_population: PopulationType = []
            for i in range(len(offspring_population)):
                individual_to_mutate = offspring_population[i]  # Ya es una copia
                try:
                    mutated_individual = self.mutation_func(
                        individual_to_mutate, **self.mutation_args  # Pasar la copia
                    )
                except Exception as e:  # Captura errores de mutación
                    str_error = "Error durante la mutación en la generación"
                    str_error += f" {generation + 1}: {e}"
                    raise GeneticAlgorithmError(str_error) from e  # type: ignore
                mutated_offspring_population.append(mutated_individual)

            # Formar la nueva población
            self.population = elite_individuals + mutated_offspring_population
            # Asegurar que la población tenga el tamaño correcto,
            # truncando si es necesario
            self.population = self.population[: self.population_size]

        print("\nOptimización Finalizada.")
        if self.best_solution_overall is not None:
            print(f"Mejor fitness global encontrado: {self.best_fitness_overall:.4f}")
        else:
            print(
                "No se encontró ninguna solución (la población podría haber colapsado)."
            )

        return self.best_solution_overall, self.best_fitness_overall

    def plot_fitness(self, save_path: Optional[str] = None) -> None:
        """
        Grafica la evolución del mejor fitness a lo largo de las generaciones.

        Utiliza Matplotlib para generar la gráfica. Si Matplotlib no está
        instalado, imprime un mensaje de advertencia.

        Args:
            save_path: Ruta opcional para guardar la gráfica como archivo de imagen.
                       Ej: "fitness_evolution.png". Si es None, solo muestra la gráfica.
        """
        if not self.best_solutions_fitness:
            print("No hay datos de fitness para graficar (ejecute el AG primero).")
            return

        try:
            import matplotlib.pyplot as plt

            plt.figure(figsize=(10, 6))
            plt.plot(
                range(1, len(self.best_solutions_fitness) + 1),  # Generaciones desde 1
                self.best_solutions_fitness,
                marker="o",
                linestyle="-",
                markersize=4,
                color="b",
                label="Mejor Fitness por Generación",
            )
            plt.axhline(
                y=self.best_fitness_overall,
                color="r",
                linestyle="--",
                label=f"Mejor Fitness Global: {self.best_fitness_overall:.4f}",
            )

            plt.title("Evolución del Fitness por Generación")
            plt.xlabel("Generación")
            plt.ylabel("Mejor Fitness")
            plt.legend()
            plt.grid(True, linestyle="--", alpha=0.7)
            plt.tight_layout()

            if save_path:
                plt.savefig(save_path)
                print(f"Gráfica de fitness guardada en: {save_path}")
            else:
                plt.show()
        except ImportError:
            print(
                "Matplotlib no está instalado. No se puede graficar. "
                "Instálalo con: pip install matplotlib"
            )
        except Exception as e:
            print(f"Ocurrió un error al graficar el fitness: {e}")



"""Módulo que contiene excepciones personalizadas para PySAG.

Este módulo define excepciones específicas para manejar errores en algoritmos genéticos.
"""

from typing import Any, Dict, List, Optional, Type, TypeVar, Union, get_origin

# Tipo genérico para los valores de los errores
T = TypeVar("T")


class PySAGError(Exception):
    """Clase base para todas las excepciones de PySAG.

    Args:
        message: Mensaje descriptivo del error.
        details: Información adicional sobre el error.

    """

    def __init__(self, message: str, details: Optional[Dict[str, Any]] = None):
        """Inicializa la excepción con un mensaje y detalles opcionales."""
        self.message = message
        self.details = details or {}
        super().__init__(self.message)

    def __str__(self) -> str:
        """Representación en cadena de la excepción."""
        if self.details:
            return f"{self.message} | Detalles: {self.details}"
        return self.message


class GeneticAlgorithmError(PySAGError):
    """Excepción base para errores en el algoritmo genético."""

    pass


class InitializationError(GeneticAlgorithmError):
    """Se produce cuando hay un error en la inicialización de la población."""

    pass


class SelectionError(GeneticAlgorithmError):
    """Se produce cuando hay un error en la selección de individuos."""

    pass


class CrossoverError(GeneticAlgorithmError):
    """Se produce cuando hay un error en el operador de cruce."""

    pass


class MutationError(GeneticAlgorithmError):
    """Se produce cuando hay un error en el operador de mutación."""

    pass


class FitnessEvaluationError(GeneticAlgorithmError):
    """Se produce cuando hay un error en la evaluación de la función de aptitud."""

    pass


class ParameterError(PySAGError):
    """Se produce cuando hay un error en los parámetros de entrada."""

    pass


class ValidationError(PySAGError):
    """Se produce cuando falla la validación de un valor o parámetro.

    Args:
        param_name: Nombre del parámetro que falló la validación.
        param_value: Valor que falló la validación.
        expected: Descripción del valor esperado.
    """  # D205, D400 corregidos aquí

    def __init__(
        self, param_name: str, param_value: Any, expected: str, **kwargs: Any
    ) -> None:
        """
        Inicializa la excepción.

        Args:
            param_name: Nombre del parámetro que falló la validación.
            param_value: Valor que falló la validación.
            expected: Descripción del valor esperado.

        """
        message = (
            f"Validación fallida para el parámetro '{param_name}'. "
            f"Se recibió: {param_value}. Se esperaba: {expected}"
        )
        details = {
            "param_name": param_name,
            "param_value": str(param_value),
            "expected": expected,
            **kwargs,
        }
        super().__init__(message, details)


class RangeError(ValidationError):
    """Se produce cuando un valor está fuera del rango permitido."""

    def __init__(
        self,
        param_name: str,
        param_value: Union[int, float],
        min_val: Optional[Union[int, float]] = None,
        max_val: Optional[Union[int, float]] = None,
        inclusive: bool = True,
        details: Optional[str] = None,
    ) -> None:
        """
        Inicializa la excepción.

        Args:
            param_name: Nombre del parámetro que falló la validación.
            param_value: Valor que falló la validación.
            min_val: Valor mínimo permitido (inclusive).
            max_val: Valor máximo permitido (inclusive).
            inclusive: Si True, el rango es inclusivo.
            details: Información adicional sobre el error.
        """
        range_desc_parts = []
        if min_val is not None:
            range_desc_parts.append(f"{'>=' if inclusive else '>'} {min_val}")
        if max_val is not None:
            range_desc_parts.append(f"{'<=' if inclusive else '<'} {max_val}")

        range_desc = (
            " y ".join(range_desc_parts) if range_desc_parts else "un valor válido"
        )
        expected = f"valor en el rango: {range_desc}"
        if details:
            expected += f" ({details})"

        super().__init__(
            param_name=param_name,
            param_value=param_value,
            expected=expected,
            min_val=min_val,
            max_val=max_val,
            inclusive=inclusive,
        )


def _format_expected_type(expected_type_val: Any) -> str:
    """Formatea un tipo esperado a una cadena legible."""
    origin = get_origin(expected_type_val)
    if origin:  # Es un tipo genérico como List[int], NDArray[float]
        return (
            str(expected_type_val).replace("typing.", "").replace("numpy.typing.", "")
        )
    elif hasattr(expected_type_val, "__name__"):
        return expected_type_val.__name__
    else:
        return str(expected_type_val)


class TypeValidationError(ValidationError):
    """Se produce cuando un valor tiene un tipo incorrecto."""

    def __init__(
        self,
        param_name: str,
        param_value: Any,
        expected_type: Union[Type, List[Any], Any],
    ) -> None:
        """
        Inicializa la excepción.

        Args:
            param_name: Nombre del parámetro que falló la validación.
            param_value: Valor que falló la validación.
            expected_type: Tipo esperado o lista de tipos esperados.
        """
        actual_type_name = type(param_value).__name__

        if isinstance(expected_type, list):
            expected_str = "uno de los tipos:\n" + ", ".join(
                [_format_expected_type(t) for t in expected_type]
            )
        else:
            expected_str = f"tipo {_format_expected_type(expected_type)}"

        super().__init__(
            param_name=param_name,
            param_value=f"(tipo: {actual_type_name})",
            expected=expected_str,
            actual_type=actual_type_name,
        )


def validate_parameter(
    value: T,
    name: str,
    expected_type: Union[Type, List[Any], Any],
    min_val: Optional[Union[int, float]] = None,
    max_val: Optional[Union[int, float]] = None,
    inclusive: bool = True,
) -> T:
    """
    Función de validación de parámetros.

    Args:
        value: Valor a validar.
        name: Nombre del parámetro.
        expected_type: Tipo esperado o lista de tipos esperados.
        min_val: Valor mínimo permitido (inclusive).
        max_val: Valor máximo permitido (inclusive).
        inclusive: Si True, el rango es inclusivo.

    Returns:
        El valor validado.

    Raises:
        TypeValidationError: Si el valor no cumple con el tipo esperado.
        RangeError: Si el valor está fuera del rango permitido.

    """
    raw_expected_types = (
        [expected_type] if not isinstance(expected_type, list) else expected_type
    )

    instance_check_types = []
    for et in raw_expected_types:
        origin = get_origin(et)
        if origin is not None:
            instance_check_types.append(origin)
        else:
            instance_check_types.append(et)

    if not instance_check_types or not any(
        isinstance(value, t) for t in instance_check_types if t is not Any
    ):  # type: ignore
        # Si t es Any, isinstance(value, Any) no es lo que queremos.
        # Any debería permitir cualquier tipo,
        # así que si Any está en los tipos esperados, pasa.
        # Esta condición se vuelve más compleja si 'Any'
        # es uno de los múltiples tipos esperados.
        # Por simplicidad, si Any es un expected_type,
        # se asume que la validación de tipo pasa.
        is_any_expected = any(et is Any for et in raw_expected_types)  # type: ignore

        if not is_any_expected:
            raise TypeValidationError(name, value, expected_type)

    # Validación de rango para números
    if isinstance(value, (int, float)) and (min_val is not None or max_val is not None):
        if min_val is not None:
            if inclusive and value < min_val:
                raise RangeError(
                    name, value, min_val=min_val, max_val=max_val, inclusive=inclusive
                )
            if not inclusive and value <= min_val:
                raise RangeError(
                    name, value, min_val=min_val, max_val=max_val, inclusive=inclusive
                )
        if max_val is not None:
            if inclusive and value > max_val:
                raise RangeError(
                    name, value, min_val=min_val, max_val=max_val, inclusive=inclusive
                )
            if not inclusive and value >= max_val:
                raise RangeError(
                    name, value, min_val=min_val, max_val=max_val, inclusive=inclusive
                )
    return value
